import glob
import os
from multiprocessing import Pool
from monty.serialization import dumpfn, loadfn

from apex.core.calculator.calculator import make_calculator
from apex.core.property.Elastic import Elastic
from apex.core.property.EOS import EOS
from apex.core.property.Gamma import Gamma
from apex.core.property.Interstitial import Interstitial
from apex.core.property.Surface import Surface
from apex.core.property.Vacancy import Vacancy
from apex.core.property.Phonon import Phonon
from apex.core.lib.utils import create_path
from apex.core.lib.util import collect_task
from apex.core.lib.dispatcher import make_submission
from apex.utils import sepline, get_task_type, handle_prop_suffix
from dflow.python import upload_packages
upload_packages.append(__file__)

lammps_task_type = ['deepmd', 'eam_alloy', 'meam', 'eam_fs', 'meam_spline', 'snap', 'gap', 'rann', 'mace']

def make_property_instance(parameters, inter_param):
    """
    Make an instance of Property
    """
    prop_type = parameters["type"]
    if prop_type == "eos":
        return EOS(parameters, inter_param)
    elif prop_type == "elastic":
        return Elastic(parameters, inter_param)
    elif prop_type == "vacancy":
        return Vacancy(parameters, inter_param)
    elif prop_type == "interstitial":
        return Interstitial(parameters, inter_param)
    elif prop_type == "surface":
        return Surface(parameters, inter_param)
    elif prop_type == "gamma":
        return Gamma(parameters, inter_param)
    elif prop_type == "phonon":
        return Phonon(parameters, inter_param)
    else:
        raise RuntimeError(f"unknown APEX type {prop_type}")


def make_property(confs, inter_param, property_list):
    # find all POSCARs and their name like mp-xxx
    # conf_dirs = glob.glob(confs)
    # conf_dirs.sort()
    conf_dirs = []
    for conf in confs:
        conf_dirs.extend(glob.glob(conf))
    conf_dirs = list(set(conf_dirs))
    conf_dirs.sort()
    for ii in conf_dirs:
        sepline(ch=ii, screen=True)
        path_to_equi = os.path.join(ii, "relaxation", "relax_task")
        try:
            structure_dict = loadfn(os.path.join(path_to_equi, "structure.json"))
        except FileNotFoundError:
            structure_dict = {}
        mismatch = structure_dict.get("mismatch", False)
        for jj in property_list:
            do_refine, suffix = handle_prop_suffix(jj)
            if not suffix:
                continue
            # generate working directory like mp-xxx/eos_00 if jj['type'] == 'eos'
            # handel the exception that the working directory exists
            # determine the suffix: from scratch or refine

            property_type = jj["type"]
            path_to_work = os.path.join(ii, property_type + "_" + suffix)
            skip_mismatch = jj.get("skip_mismatch", False)
            if mismatch and skip_mismatch:
                print("Skip mismatched structure")
                continue

            create_path(path_to_work)

            inter_param_prop = jj.get("cal_setting", {}).get("overwrite_interaction", inter_param)

            prop = make_property_instance(jj, inter_param_prop)
            task_list = prop.make_confs(path_to_work, path_to_equi, do_refine)

            for kk in task_list:
                poscar = os.path.join(kk, "POSCAR")
                inter = make_calculator(inter_param_prop, poscar)
                inter.make_potential_files(kk)
                inter.make_input_file(kk, prop.task_type(), prop.task_param())

            prop.post_process(task_list)  # generate same KPOINTS file for elastic when doing DFT


def worker(
    work_path,
    all_task,
    forward_common_files,
    forward_files,
    backward_files,
    mdata,
    inter_type,
    task_type,
):
    run_tasks = [os.path.basename(ii) for ii in all_task]
    machine = mdata.get("machine", None)
    resources = mdata.get("resources", None)
    command = mdata.get(f"{task_type}_run_command", mdata.get("run_command", None))
    group_size = mdata.get("group_size", 1)

    submission = make_submission(
        mdata_machine=machine,
        mdata_resources=resources,
        commands=[command],
        work_path=work_path,
        run_tasks=run_tasks,
        group_size=group_size,
        forward_common_files=forward_common_files,
        forward_files=forward_files,
        backward_files=backward_files,
        outlog="outlog",
        errlog="errlog",
    )
    submission.run_submission()


def run_property(confs, inter_param, property_list, mdata):
    # find all POSCARs and their name like mp-xxx
    # conf_dirs = glob.glob(confs)
    # conf_dirs.sort()
    conf_dirs = []
    for conf in confs:
        conf_dirs.extend(glob.glob(conf))
    conf_dirs = list(set(conf_dirs))
    conf_dirs.sort()

    task_list = []
    work_path_list = []
    multiple_ret = []
    for ii in conf_dirs:
        sepline(ch=ii, screen=True)
        for jj in property_list:
            # determine the suffix: from scratch or refine
            # ...
            do_refine, suffix = handle_prop_suffix(jj)
            if not suffix:
                continue

            property_type = jj["type"]
            path_to_work = os.path.abspath(
                os.path.join(ii, property_type + "_" + suffix)
            )

            work_path_list.append(path_to_work)
            tmp_task_list = glob.glob(os.path.join(path_to_work, "task.[0-9]*[0-9]"))
            tmp_task_list.sort()
            task_list.append(tmp_task_list)

            inter_param_prop = jj.get("cal_setting", {}).get("overwrite_interaction", inter_param)

            # dispatch the tasks
            # POSCAR here is useless
            virtual_calculator = make_calculator(inter_param_prop, "POSCAR")
            forward_files = virtual_calculator.forward_files(property_type)
            forward_common_files = virtual_calculator.forward_common_files(
                property_type
            )
            backward_files = virtual_calculator.backward_files(property_type)
            #    backward_files += logs
            # ...
            task_type = get_task_type({"interaction": inter_param})
            inter_type = inter_param_prop["type"]
            work_path = path_to_work
            all_task = tmp_task_list
            run_tasks = collect_task(all_task, inter_type)
            if len(run_tasks) == 0:
                continue
            else:
                processes = len(run_tasks)
                pool = Pool(processes=processes)
                print("Submit job via %d processes" % processes)
                ret = pool.apply_async(
                    worker,
                    (
                        work_path,
                        all_task,
                        forward_common_files,
                        forward_files,
                        backward_files,
                        mdata,
                        inter_type,
                        task_type
                    )
                )
                multiple_ret.append(ret)
    pool.close()
    pool.join()
    for ii in range(len(multiple_ret)):
        if not multiple_ret[ii].successful():
            print("ERROR:", multiple_ret[ii].get())
            raise RuntimeError("Job %d is not successful!" % ii)
    print("%d jobs are finished" % len(multiple_ret))


def post_property(confs, inter_param, property_list):
    # find all POSCARs and their name like mp-xxx
    #    task_list = []
    # conf_dirs = glob.glob(confs)
    # conf_dirs.sort()
    conf_dirs = []
    for conf in confs:
        conf_dirs.extend(glob.glob(conf))
    conf_dirs = list(set(conf_dirs))
    conf_dirs.sort()
    for ii in conf_dirs:
        for jj in property_list:
            # determine the suffix: from scratch or refine
            # ...
            do_refine, suffix = handle_prop_suffix(jj)
            if not suffix:
                continue

            inter_param_prop = jj.get("cal_setting", {}).get("overwrite_interaction", inter_param)

            property_type = jj["type"]
            path_to_work = os.path.join(ii, property_type + "_" + suffix)
            prop = make_property_instance(jj, inter_param_prop)
            param_json = os.path.join(path_to_work, "param.json")
            param_dict = prop.parameter
            param_dict.setdefault("skip", False) # default of "skip" is False
            try:
                param_dict.pop("skip")
            except KeyError:
                pass
            dumpfn(param_dict, param_json)
            prop.compute(
                os.path.join(path_to_work, "result.json"),
                os.path.join(path_to_work, "result.out"),
                path_to_work
            )



import logging
import datetime
from pymongo import MongoClient
from bson.json_util import dumps, loads
from apex.database.StorageBase import StorageBase
from apex.utils import update_dict


class MongoDBClient(StorageBase):
    def __init__(
        self,
        name: str,
        database_name: str,
        collection_name: str,
        **kwargs
    ):
        super().__init__(name)
        self.client = MongoClient(**kwargs)
        # Send a ping to confirm a successful connection
        try:
            self.client.admin.command('ping')
            logging.info(msg="Successfully connected to MongoDB!")
        except Exception as e:
            raise e
        self.db = self.client[database_name]
        self.collection = self.db[collection_name]

    def sync(self, data: dict, id_field: str, depth: int = 9999):
        """synchronize dict data to MongoDB"""
        logging.info(msg=f'Synchronize data into MongoDB {self.collection}')
        if self.collection.count_documents({'_id': id_field}, limit=1) != 0:
            logging.info(msg=f'Synchronizing with exist dataset (_id: {id_field})')
            orig_dict = self.collection.find_one({'_id': id_field})
            update_dict(orig_dict, data, depth)
            self.collection.update_one({'_id': id_field}, {"$set": orig_dict})
        else:
            logging.info(msg=f'Creating new dataset... (_id: {id_field})')
            self.collection.insert_one(data)

    def record(self, data: dict, id_field: str):
        """record dict data to MongoDB"""
        logging.info(msg=f'Record data into MongoDB {self.collection}')
        # get timestamp
        timestamp = datetime.datetime.now().isoformat()
        _id = f'[{timestamp}]:{id_field}'
        logging.info(msg=f'Creating new dataset... (_id: {_id})')
        data['_id'] = _id
        self.collection.insert_one(data)

    def load_json(self, query):
        """load BSON from MongoDB"""
        cursor = self.collection.find(query)
        return [loads(dumps(doc)) for doc in cursor]

    def close(self):
        self.client.close()



#!/usr/bin/env python3
import logging
import os
import shutil
import json
import string
import random
from typing import Type, List
from monty.serialization import loadfn
from decimal import Decimal
from dflow.python import OP
from dflow.python import upload_packages
from fpop.vasp import RunVasp
from fpop.abacus import RunAbacus
from apex.op.RunLAMMPS import RunLAMMPS
from apex.core.calculator import LAMMPS_INTER_TYPE

upload_packages.append(__file__)

MaxLength = 70
# LAMMPS_INTER_TYPE = ['deepmd', 'eam_alloy', 'meam', 'eam_fs', 'meam_spline', 'snap', 'gap', 'rann', 'mace']


# write a function to replace all '/' in the input string with '-'

def backup_path(path) -> None:
    path += "/"
    if os.path.isdir(path):
        dirname = os.path.dirname(path)
        counter = 0
        while True:
            bk_dirname = dirname + ".bk%03d" % counter
            if not os.path.isdir(bk_dirname):
                shutil.move(dirname, bk_dirname)
                break
            counter += 1


def generate_random_string(length):
    characters = string.ascii_letters + string.digits  # 包含所有字母(大写和小写)和数字
    random_string = ''.join(random.choices(characters, k=length))
    return random_string


def copy_all_other_files(
        src_dir,
        dst_dir,
        exclude_files=[],
        include_dirs=[]
) -> None:
    """
    Copies all files from the source directory to the destination directory with some files excluded
    and some directories included.

    :param src_dir: The path to the source directory.
    :param dst_dir: The path to the destination directory.
    :exclude_files: files to be ignored.
    :include_dirs: directories to be included.
    """
    if not os.path.exists(src_dir):
        raise FileNotFoundError(f"Source directory {src_dir} does not exist.")

    if not os.path.exists(dst_dir):
        os.makedirs(dst_dir)

    for item in os.listdir(src_dir):
        src_path = os.path.join(src_dir, item)
        dst_path = os.path.join(dst_dir, item)

        if os.path.isfile(src_path) and item not in exclude_files:
            shutil.copy2(src_path, dst_path)
        elif os.path.isdir(src_path) and item in include_dirs:
            shutil.copytree(src_path, dst_path)


def simplify_paths(path_list: list) -> dict:
    # only one path, return it with only basename
    if len(path_list) == 1:
        return {path_list[0]: '.../' + os.path.basename(path_list[0])}
    else:
        # Split all paths into components
        split_paths = [os.path.normpath(p).split(os.sep) for p in path_list]

        # Find common prefix
        common_prefix = os.path.commonprefix(split_paths)
        common_prefix_len = len(common_prefix)

        # Remove common prefix from each path and create dictionary
        simplified_paths_dict = {
            os.sep.join(p): '.../' + os.sep.join(p[common_prefix_len:]) if common_prefix_len else os.sep.join(p)
            for p in split_paths
        }

        return simplified_paths_dict


def is_json_file(filename):
    try:
        with open(filename, 'r') as f:
            json.load(f)
        return True
    except ValueError as e:
        return False


def load_config_file(config_file: os.PathLike) -> dict:
    try:
        config_dict = loadfn(config_file)
    except FileNotFoundError:
        logging.warning(
            msg='No global config file provided, will default all settings. '
                'You may prepare global.json under current work direction '
                'or use optional argument: -c to indicate a specific json file.'
        )
        config_dict = {}
    return config_dict


def recursive_search(directories, path='.'):
    """recursive search target directory"""
    # list all directions
    items = os.listdir(path)
    directories_in_path = [
        i for i in items if os.path.isdir(os.path.join(path, i)) and not i.startswith('.')
    ]

    # check if target work direction is found
    if set(directories) <= set(directories_in_path):
        return os.path.abspath(path)

    # recursive search in next direction
    if len(directories_in_path) == 1:
        return recursive_search(directories, os.path.join(path, directories_in_path[0]))

    # return False for failure
    return False


def handle_prop_suffix(parameter: dict):
    if parameter.get('skip', False):
        return None, None
    if 'init_from_suffix' and 'output_suffix' in parameter:
        do_refine = True
        suffix = parameter['output_suffix']
    elif 'reproduce' in parameter and parameter['reproduce']:
        do_refine = False
        suffix = 'reprod'
    elif 'suffix' in parameter and parameter['suffix']:
        do_refine = False
        suffix = str(parameter['suffix'])
    else:
        do_refine = False
        suffix = '00'
    return do_refine, suffix


def return_prop_list(parameters: list) -> list:
    prop_list = []
    for ii in parameters:
        _, suffix = handle_prop_suffix(ii)
        if not suffix:
            continue
        prop_list.append(ii['type'] + '_' + suffix)
    return prop_list


def get_flow_type(d: dict) -> str:
    if 'relaxation' in d and 'properties' not in d:
        flow_type = 'relax'
    elif 'properties' in d and 'relaxation' not in d:
        flow_type = 'props'
    elif 'relaxation' in d and 'properties' in d:
        flow_type = 'joint'
    else:
        raise RuntimeError('Can not recognize type of the input json file')
    return flow_type


def get_task_type(d: dict) -> (str, Type[OP]):
    interaction_type = d['interaction']['type']
    if interaction_type == 'vasp':
        task_type = 'vasp'
        run_op = RunVasp
    elif interaction_type == 'abacus':
        task_type = 'abacus'
        run_op = RunAbacus
    elif interaction_type in LAMMPS_INTER_TYPE:
        task_type = 'lammps'
        run_op = RunLAMMPS
    else:
        raise RuntimeError(f'Unsupported interaction type: {interaction_type}')

    return task_type, run_op


def judge_flow(parameter: List[dict], specify: str) -> (Type[OP], str, str, dict, dict):
    # identify type of flow and input parameter file
    num_args = len(parameter)
    if num_args == 1:
        task, run_op = get_task_type(parameter[0])
        flow = get_flow_type(parameter[0])
        task_type = task
        if flow == 'relax':
            flow_type = 'relax'
            if specify in ['props', 'joint']:
                raise RuntimeError(
                    'relaxation json file argument provided! Please check your jason file.'
                )
            relax_param = parameter[0]
            props_param = None
        elif flow == 'props':
            if specify in ['relax', 'joint']:
                raise RuntimeError(
                    'property test json file argument provided! Please check your jason file.'
                )
            flow_type = 'props'
            relax_param = None
            props_param = parameter[0]
        else:
            if specify == 'relax':
                flow_type = 'relax'
            elif specify == 'props':
                flow_type = 'props'
            else:
                flow_type = 'joint'
            relax_param = parameter[0]
            props_param = parameter[0]

    elif num_args == 2:
        task1, run_op1 = get_task_type(parameter[0])
        flow1 = get_flow_type(parameter[0])
        task2, run_op2 = get_task_type(parameter[1])
        flow2 = get_flow_type(parameter[1])
        if flow1 != flow2:
            if specify == 'relax':
                flow_type = 'relax'
            elif specify == 'props':
                flow_type = 'props'
            else:
                flow_type = 'joint'
            if flow1 == 'relax' and flow2 == 'props':
                relax_param = parameter[0]
                props_param = parameter[1]
            elif flow1 == 'props' and flow2 == 'relax':
                relax_param = parameter[1]
                props_param = parameter[0]
            else:
                raise RuntimeError(
                    'confusion of json arguments provided: '
                    'joint type of json conflicts with the other json argument'
                )
        else:
            raise RuntimeError('Same type of input json files')
        if task1 == task2:
            task_type = task1
            run_op = run_op1
        else:
            raise RuntimeError('interaction types given are not matched')
    else:
        raise ValueError('A maximum of two input arguments is allowed')

    return run_op, task_type, flow_type, relax_param, props_param


def sepline(ch="-", sp="-", screen=False):
    r"""
    seperate the output by '-'
    """
    ch.center(MaxLength, sp)


def update_dict(d_base: dict, d_new: dict, depth=9999) -> None:
    depth -= 1
    if d_new is None:
        return None
    for k, v in d_new.items():
        if isinstance(v, dict) and k in d_base and isinstance(d_base[k], dict) and depth >= 0:
            update_dict(d_base[k], v, depth)
        else:
            d_base[k] = v


def convert_floats_to_decimals(obj):
    if isinstance(obj, float):
        return Decimal(str(obj))
    elif isinstance(obj, dict):
        return {k: convert_floats_to_decimals(v) for k, v in obj.items()}
    elif isinstance(obj, (list, tuple)):
        return [convert_floats_to_decimals(x) for x in obj]
    else:
        return obj


def json2dict(function):
    def wrapper(*args, **kwargs):
        # try to convert json to dict for arguments passed as args
        args = list(args)
        for ii in range(len(args)):
            if isinstance(args[ii], os.PathLike) or isinstance(args[ii], str):
                try:
                    args[ii] = loadfn(args[ii])
                except Exception:
                    pass
        # try to convert json to dict for arguments passed as kwargs
        for k, v in kwargs.items():
            if isinstance(v, os.PathLike) or isinstance(v, str):
                try:
                    kwargs[k] = loadfn(v)
                except Exception:
                    pass
        result = function(*tuple(args), **kwargs)
        return result
    return wrapper



import glob
import json
import logging
import os
import re
import shutil

import numpy as np
from monty.serialization import dumpfn, loadfn
from pymatgen.analysis.defects.generators import VoronoiInterstitialGenerator
from pymatgen.analysis.defects.core import Interstitial as pmgInterstitial
from pymatgen.core.structure import Structure, Lattice
from pymatgen.core.sites import PeriodicSite
from pymatgen.core.tensors import Tensor
from pymatgen.core.operations import SymmOp

from apex.core.calculator.lib import abacus_utils
from apex.core.calculator.lib import lammps_utils
from apex.core.property.Property import Property
from apex.core.refine import make_refine
from apex.core.reproduce import make_repro, post_repro
from apex.core.structure import StructureInfo
from dflow.python import upload_packages

upload_packages.append(__file__)

PREDEFINED_LIST = ['bcc', 'fcc', 'hcp']
TOL = 1e-5

class Interstitial(Property):
    def __init__(self, parameter, inter_param=None):
        parameter["reproduce"] = parameter.get("reproduce", False)
        self.reprod = parameter["reproduce"]
        if not self.reprod:
            if not ("init_from_suffix" in parameter and "output_suffix" in parameter):
                default_supercell = [1, 1, 1]
                parameter["supercell"] = parameter.get("supercell", default_supercell)
                self.supercell = parameter["supercell"]
                self.insert_ele = parameter.get("insert_ele", None)
                parameter["lattice_type"] = parameter.get("lattice_type", None)
                self.lattice_type = parameter["lattice_type"]
                parameter["voronoi_param"] = parameter.get("voronoi_param", {})
                self.voronoi_param = parameter["voronoi_param"]
            parameter["cal_type"] = parameter.get("cal_type", "relaxation")
            default_cal_setting = {
                "relax_pos": True,
                "relax_shape": True,
                "relax_vol": True,
            }
        else:
            parameter["cal_type"] = "static"
            default_cal_setting = {
                "relax_pos": False,
                "relax_shape": False,
                "relax_vol": False,
            }
            parameter["init_from_suffix"] = parameter.get("init_from_suffix", "00")
            self.init_from_suffix = parameter["init_from_suffix"]
        self.cal_type = parameter["cal_type"]
        parameter["cal_setting"] = parameter.get("cal_setting", default_cal_setting)
        for key in default_cal_setting:
            parameter["cal_setting"].setdefault(key, default_cal_setting[key])
        self.cal_setting = parameter["cal_setting"]
        self.parameter = parameter
        self.inter_param = inter_param if inter_param != None else {"type": "vasp"}

    def make_confs(self, path_to_work, path_to_equi, refine=False):
        self.path_to_work = os.path.abspath(path_to_work)
        if os.path.exists(path_to_work):
            logging.debug("%s already exists" % path_to_work)
        else:
            os.makedirs(path_to_work)
        path_to_equi = os.path.abspath(path_to_equi)

        if "start_confs_path" in self.parameter and os.path.exists(
                self.parameter["start_confs_path"]
        ):
            init_path_list = glob.glob(
                os.path.join(self.parameter["start_confs_path"], "*")
            )
            struct_init_name_list = [os.path.basename(ii) for ii in init_path_list]
            struct_output_name = os.path.basename(os.path.dirname(path_to_work))
            assert struct_output_name in struct_init_name_list, f"{struct_output_name} not in initial configurations"
            path_to_equi = os.path.abspath(
                os.path.join(
                    self.parameter["start_confs_path"],
                    struct_output_name,
                    "relaxation",
                    "relax_task",
                )
            )

        self.task_list = []
        cwd = os.getcwd()

        if self.reprod:
            print("interstitial reproduce starts")
            if "init_data_path" not in self.parameter:
                raise RuntimeError("please provide the initial data path to reproduce")
            init_data_path = os.path.abspath(self.parameter["init_data_path"])
            self.task_list = make_repro(
                self.inter_param,
                init_data_path,
                self.init_from_suffix,
                self.path_to_work,
                self.parameter.get("reprod_last_frame", False),
            )

        else:
            if refine:
                print("interstitial refine starts")
                self.task_list = make_refine(
                    self.parameter["init_from_suffix"],
                    self.parameter["output_suffix"],
                    self.path_to_work,
                )

                init_from_path = re.sub(
                    self.parameter["output_suffix"][::-1],
                    self.parameter["init_from_suffix"][::-1],
                    self.path_to_work[::-1],
                    count=1,
                )[::-1]
                task_list_basename = list(map(os.path.basename, self.task_list))

                os.chdir(self.path_to_work)
                if os.path.exists("element.out"):
                    os.remove("element.out")
                os.symlink(
                    os.path.relpath(os.path.join(init_from_path, "element.out")),
                    "element.out",
                )
                os.chdir(cwd)

                for ii in task_list_basename:
                    init_from_task = os.path.join(init_from_path, ii)
                    output_task = os.path.join(self.path_to_work, ii)
                    os.chdir(output_task)
                    if os.path.exists("supercell.json"):
                        os.remove("supercell.json")
                    os.symlink(
                        os.path.relpath(os.path.join(init_from_task, "supercell.json")),
                        "supercell.json",
                    )

            else:
                if self.inter_param["type"] == "abacus":
                    CONTCAR = abacus_utils.final_stru(path_to_equi)
                    POSCAR = "STRU"
                else:
                    CONTCAR = "CONTCAR"
                    POSCAR = "POSCAR"

                equi_contcar = os.path.join(path_to_equi, CONTCAR)
                orig_poscar = os.path.join(path_to_equi, POSCAR)
                if not os.path.exists(equi_contcar):
                    raise RuntimeError("please do relaxation first")

                if self.inter_param["type"] == "abacus":
                    ss = abacus_utils.stru2Structure(equi_contcar)
                else:
                    ss = Structure.from_file(equi_contcar)
                rot = Tensor.get_ieee_rotation(ss)
                op = SymmOp.from_rotation_and_translation(rot)
                ss.apply_operation(op)
                # get structure type
                os.chdir(self.path_to_work)
                # convert site element into same type for a pseudo-structure just for simple lattice type judgment
                same_type_ss = ss.copy()
                species_mapping = {str(specie): "Ni" for specie in same_type_ss.composition.elements}
                same_type_ss.replace_species(species_mapping)
                st = StructureInfo(same_type_ss, symprec=0.1, angle_tolerance=5)
                # indication of structure type
                self.structure_type = st.lattice_structure
                # get conventional standard structure to ss
                orig_st = StructureInfo(ss, symprec=0.1, angle_tolerance=5)
                conv_ss = orig_st.conventional_structure
                conv_ss.to("POSCAR", "POSCAR")
                conv_ss.to("POSCAR_conv", "POSCAR")
                ss = conv_ss
                if self.lattice_type:
                    print(f'Adopt user indicated lattice type: {self.lattice_type}')
                    self.structure_type = self.lattice_type
                os.chdir(cwd)

                # gen defects
                dss = []
                self.insert_element_task = os.path.join(self.path_to_work, "element.out")
                if os.path.isfile(self.insert_element_task):
                    os.remove(self.insert_element_task)
                if not self.insert_ele:
                    self.insert_ele = [str(ii) for ii in set(ss.composition.elements)]
                for ii in self.insert_ele:
                    if self.structure_type in PREDEFINED_LIST:
                        # rotate and translate hcp structure to specific orientation for interstitial generation
                        if self.structure_type == 'hcp':
                            theta = -2 * np.pi / 3
                            rot_m = np.array([
                                [np.cos(theta), -np.sin(theta), 0],
                                [np.sin(theta), np.cos(theta), 0],
                                [0, 0, 1]
                            ])
                            op = SymmOp.from_rotation_and_translation(rotation_matrix=rot_m)
                            ss.apply_operation(op)
                            new_lattice = Lattice([
                                ss.lattice.matrix[0] * -1, ss.lattice.matrix[1] * -1, ss.lattice.matrix[2]
                            ])
                            new_frac_coords = ss.frac_coords.copy()
                            if not ((new_frac_coords[0][0] < 0.5 and new_frac_coords[0][2] < 0.5)\
                                    or (new_frac_coords[0][0] > 0.5 and new_frac_coords[0][2] > 0.5)):
                                new_frac_coords[0][2] = ss.frac_coords[1][2]
                                new_frac_coords[1][2] = ss.frac_coords[0][2]
                            new_ss = Structure(new_lattice, ss.species, new_frac_coords, coords_are_cartesian=False)
                            ss = new_ss
                            ss.to(os.path.join(self.path_to_work, 'POSCAR_conv'), 'POSCAR')
                        # produce a pseudo interstitial structure for later modification
                        vds = [pmgInterstitial(ss, PeriodicSite(ii, [0.12, 0.13, 0.14], ss.lattice))]
                    else:
                        pre_vds = VoronoiInterstitialGenerator(**self.voronoi_param)
                        vds = pre_vds.generate(ss, [ii])
                    for jj in vds:
                        temp = jj.get_supercell_structure(
                            sc_mat=np.diag(self.supercell, k=0)
                        )
                        smallest_distance = list(set(temp.distance_matrix.ravel()))[1]
                        if (
                                "conf_filters" in self.parameter
                                and "min_dist" in self.parameter["conf_filters"]
                        ):
                            min_dist = self.parameter["conf_filters"]["min_dist"]
                            if smallest_distance >= min_dist:
                                dss.append(temp)
                                with open(self.insert_element_task, "a+") as fout:
                                    print(ii, file=fout)
                        else:
                            dss.append(temp)
                            with open(self.insert_element_task, "a+") as fout:
                                print(ii, file=fout)
                        #            dss.append(jj.generate_defect_structure(self.supercell))
                        self.dss = dss

                print(
                    "gen interstitial with supercell "
                    + str(self.supercell)
                    + " with element "
                    + str(self.insert_ele)
                )
                os.chdir(self.path_to_work)
                if os.path.exists(POSCAR):
                    os.remove(POSCAR)
                os.symlink(os.path.relpath(equi_contcar), POSCAR)
                for ii in range(len(dss)):
                    output_task = os.path.join(self.path_to_work, "task.%06d" % ii)
                    os.makedirs(output_task, exist_ok=True)
                    os.chdir(output_task)
                    for jj in [
                        "INCAR",
                        "POTCAR",
                        "POSCAR",
                        "conf.lmp",
                        "in.lammps",
                        "STRU",
                    ]:
                        if os.path.exists(jj):
                            os.remove(jj)
                    self.task_list.append(output_task)
                    dss[ii].to("POSCAR", "POSCAR")
                    # np.savetxt('supercell.out', self.supercell, fmt='%d')
                    dumpfn(self.supercell, "supercell.json")
                    dumpfn(f'VoronoiType_{ii}', 'interstitial_type.json')
                os.chdir(cwd)

                super_size = (
                        self.supercell[0] * self.supercell[1] * self.supercell[2]
                )
                num_atom = super_size * 2
                # chl = -num_atom - 2
                os.chdir(self.path_to_work)

                # create pre-defined special SIA structure for bcc fcc and hcp
                if self.structure_type in PREDEFINED_LIST:
                    self.task_list = []
                    if not os.path.isfile("task.000000/POSCAR"):
                        raise RuntimeError("need task.000000 structure as reference")

                    with open('POSCAR_conv', "r") as fin:
                        fin.readline()
                        scale = float(fin.readline().split()[0])
                        self.latt_param = float(fin.readline().split()[0])
                        self.latt_param *= scale

                    with open("task.000000/POSCAR", "r") as fin:
                        self.pos_line = fin.read().split("\n")

                    for idx, ii in enumerate(self.pos_line):
                        ss = ii.split()
                        if len(ss) > 3:
                            if (
                                    abs(0.12 / self.supercell[0] - float(ss[0])) < TOL
                                    and abs(0.13 / self.supercell[1] - float(ss[1])) < TOL
                                    and abs(0.14 / self.supercell[2] - float(ss[2])) < TOL
                            ):
                                chl = idx
                    shutil.rmtree("task.000000")

                    os.chdir(cwd)
                    # specify interstitial structures
                    if self.structure_type == 'bcc':
                        for idx, ii in enumerate(self.pos_line):
                            ss = ii.split()
                            if len(ss) > 3:
                                if (
                                        abs(0.5 / self.supercell[0] - float(ss[0])) < TOL
                                        and abs(0.5 / self.supercell[1] - float(ss[1])) < TOL
                                        and abs(0.5 / self.supercell[2] - float(ss[2])) < TOL
                                ):
                                    center = idx
                        bcc_interstital_dict = {
                            'tetrahedral': {chl: [0.25, 0.5, 0]},
                            'octahedral': {chl: [0.5, 0.5, 0]},
                            'crowdion': {chl: [0.25, 0.25, 0.25]},
                            '<111>dumbbell': {chl: [1 / 3, 1 / 3, 1 / 3],
                                              center: [2 / 3, 2 / 3, 2 / 3]},
                            '<110>dumbbell': {chl: [1 / 4, 3 / 4, 1 / 2],
                                              center: [3 / 4, 1 / 4, 1 / 2]},
                            '<100>dumbbell': {chl: [1 / 2, 1 / 2, 1 / 6],
                                              center: [1 / 2, 1 / 2, 5 / 6]}
                        }
                        total_task = self.__gen_tasks(bcc_interstital_dict)

                    elif self.structure_type == 'fcc':
                        for idx, ii in enumerate(self.pos_line):
                            ss = ii.split()
                            if len(ss) > 3:
                                if (
                                        abs(1 / self.supercell[0] - float(ss[0])) < TOL
                                        and abs(0.5 / self.supercell[1] - float(ss[1])) < TOL
                                        and abs(0.5 / self.supercell[2] - float(ss[2])) < TOL
                                ):
                                    face = idx

                                if (
                                        abs(1 / self.supercell[0] - float(ss[0])) < TOL
                                        and abs(1 / self.supercell[1] - float(ss[1])) < TOL
                                        and abs(1 / self.supercell[2] - float(ss[2])) < TOL
                                ):
                                    corner = idx

                        fcc_interstital_dict = {
                            'tetrahedral': {chl: [0.75, 0.25, 0.25]},
                            'octahedral': {chl: [1, 0, 0.5]},
                            'crowdion': {chl: [1, 0.25, 0.25]},
                            '<111>dumbbell': {
                                chl: [1 - 0.3 / np.sqrt(3),
                                      1 - 0.3 / np.sqrt(3),
                                      1 - 0.3 / np.sqrt(3)],
                                corner: [0.3 / np.sqrt(3),
                                         0.3 / np.sqrt(3),
                                         0.3 / np.sqrt(3)]
                            },
                            '<110>dumbbell': {
                                chl: [1,
                                      0.5 + (0.3 / np.sqrt(2)),
                                      0.5 + (0.3 / np.sqrt(2))],
                                face: [1,
                                       0.5 - (0.3 / np.sqrt(2)),
                                       0.5 - (0.3 / np.sqrt(2))]
                            },
                            '<100>dumbbell': {
                                chl: [1, 0.2, 0.5],
                                face: [1, 0.8, 0.5]
                            },
                        }
                        total_task = self.__gen_tasks(fcc_interstital_dict)

                    elif self.structure_type == 'hcp':
                        for idx, ii in enumerate(self.pos_line):
                            ss = ii.split()
                            if len(ss) > 3:
                                if (
                                        abs(1/3 / self.supercell[0] - float(ss[0])) < TOL
                                        and abs(2/3 / self.supercell[1] - float(ss[1])) < TOL
                                        and abs(0.25 / self.supercell[2] - float(ss[2])) < TOL
                                ):
                                    center = idx
                        hcp_interstital_dict = {
                            'O': {chl: [0, 0, 0.5]},
                            'BO': {chl: [0, 0, 0.25]},
                            'C': {chl: [0.5, 0.5, 0.5]},
                            'BC': {chl: [5/6, 2/3, 0.25]},
                            'S': {
                                chl: [1/3, 2/3, 0.475],
                                center: [1/3, 2/3, 0.025]
                            },
                            'BS': {
                                chl: [2/3, 2/3, 0.25],
                                center: [0, 2/3, 0.25]
                            },
                            'T': {chl: [2/3, 1/3, 0.375]},
                            'BT': {chl: [2/3, 1/3, 0.25]},
                        }
                        total_task = self.__gen_tasks(hcp_interstital_dict)
                else:
                    total_task = len(dss)

                if self.inter_param["type"] == "abacus":
                    for ii in range(total_task):
                        output_task = os.path.join(self.path_to_work, "task.%06d" % ii)
                        os.chdir(output_task)
                        abacus_utils.poscar2stru("POSCAR", self.inter_param, "STRU")
                        #os.remove("POSCAR")
        os.chdir(cwd)
        return self.task_list

    def __gen_tasks(self, interstitial_dict):
        cwd = os.getcwd()
        for ii, (type_str, adjust_dict) in enumerate(interstitial_dict.items()):
            output_task = os.path.join(
                self.path_to_work, "task.%06d" % (len(self.dss) + ii - 1)
            )
            os.makedirs(output_task, exist_ok=True)
            os.chdir(output_task)
            self.task_list.append(output_task)
            # adjust atom positions in POSCAR
            with open(self.insert_element_task, "a+") as fout:
                print(self.insert_ele[0], file=fout)
            dumpfn(self.supercell, "supercell.json")
            dumpfn(type_str, 'interstitial_type.json')
            new_pos_line = self.pos_line.copy()
            for line, pos in adjust_dict.items():
                new_pos_line[line] = (
                        "%.6f" % float(pos[0] / self.supercell[0])
                        + " "
                        + "%.6f" % float(pos[1] / self.supercell[1])
                        + " "
                        + "%.6f" % float(pos[2] / self.supercell[2])
                        + " "
                        + self.insert_ele[0]
                )
            with open("POSCAR", "w+") as fout:
                for ii in new_pos_line:
                    print(ii, file=fout)
            print(f"gen {type_str}")
            os.chdir(cwd)

        total_task = len(self.dss) + len(interstitial_dict) - 1

        return total_task


    def post_process(self, task_list):
        if True:
            fin1 = open(os.path.join(task_list[0], "..", "element.out"), "r")
            for ii in task_list:
                conf = os.path.join(ii, "conf.lmp")
                inter = os.path.join(ii, "inter.json")
                insert_ele = fin1.readline().split()[0]
                if os.path.isfile(conf):
                    with open(conf, "r") as fin2:
                        conf_line = fin2.read().split("\n")
                        insert_line = conf_line[-2]
                    type_map = loadfn(inter)["type_map"]
                    type_map_list = lammps_utils.element_list(type_map)
                    if int(insert_line.split()[1]) > len(type_map_list):
                        type_num = type_map[insert_ele] + 1
                        conf_line[2] = str(len(type_map_list)) + " atom types"
                        conf_line[-2] = (
                                "%6.d" % int(insert_line.split()[0])
                                + "%7.d" % type_num
                                + "%16.10f" % float(insert_line.split()[2])
                                + "%16.10f" % float(insert_line.split()[3])
                                + "%16.10f" % float(insert_line.split()[4])
                        )
                        with open(conf, "w+") as fout:
                            for jj in conf_line:
                                print(jj, file=fout)
            fin1.close()

    def task_type(self):
        return self.parameter["type"]

    def task_param(self):
        return self.parameter

    def _compute_lower(self, output_file, all_tasks, all_res):
        output_file = os.path.abspath(output_file)
        res_data = {}
        ptr_data = os.path.dirname(output_file) + "\n"

        if not self.reprod:
            with open(
                    os.path.join(os.path.dirname(output_file), "element.out"), "r"
            ) as fin:
                fc = fin.read().split("\n")
            ptr_data += "Insert_ele-Struct:          \tInter_E(eV)    \tE(eV)     \tequi_E(eV)\n"

            equi_path = os.path.abspath(
                os.path.join(
                    os.path.dirname(output_file), "../relaxation/relax_task"
                )
            )
            equi_result = loadfn(os.path.join(equi_path, "result.json"))
            equi_epa = equi_result["energies"][-1] / sum(equi_result["atom_numbs"])

            for idid, ii in enumerate(all_tasks, start=0): # skip task.000000
                structure_dir = os.path.basename(ii)
                task_result = loadfn(all_res[idid])
                interstitial_type = loadfn(os.path.join(ii, 'interstitial_type.json'))
                natoms = sum(task_result["atom_numbs"])
                evac = task_result["energies"][-1] - equi_epa * natoms
                supercell_index = loadfn(os.path.join(ii, "supercell.json"))
                # insert_ele = loadfn(os.path.join(ii, 'task.json'))['insert_ele'][0]
                insert_ele = fc[idid]
                ptr_data += "%s: \t%7.3f  \t%7.3f \t%7.3f \n" % (
                    insert_ele + "_" + str(interstitial_type) + "_" + structure_dir,
                    evac,
                    task_result["energies"][-1],
                    equi_epa * natoms,
                )
                res_data[
                    insert_ele + "_" + str(interstitial_type) + "_" + structure_dir
                    ] = [evac, task_result["energies"][-1], equi_epa * natoms]

        else:
            if "init_data_path" not in self.parameter:
                raise RuntimeError("please provide the initial data path to reproduce")
            init_data_path = os.path.abspath(self.parameter["init_data_path"])
            res_data, ptr_data = post_repro(
                init_data_path,
                self.parameter["init_from_suffix"],
                all_tasks,
                ptr_data,
                self.parameter.get("reprod_last_frame", False),
            )

        with open(output_file, "w") as fp:
            json.dump(res_data, fp, indent=4)

        return res_data, ptr_data



import os, glob, pathlib, shutil, subprocess, logging
from pathlib import Path
from monty.serialization import loadfn
from typing import List
from dflow.python import (
    OP,
    OPIO,
    OPIOSign,
    Artifact,
    upload_packages
)
from monty.serialization import dumpfn
from apex.utils import recursive_search
from apex.core.lib.utils import create_path
from apex.core.calculator import LAMMPS_INTER_TYPE

upload_packages.append(__file__)


class PropsMake(OP):
    """
    OP class for making calculation tasks (make property)
    """

    def __init__(self):
        pass

    @classmethod
    def get_input_sign(cls):
        return OPIOSign({
            'input_work_path': Artifact(Path),
            'path_to_prop': str,
            'prop_param': dict,
            'inter_param': dict,
            'do_refine': bool
        })

    @classmethod
    def get_output_sign(cls):
        return OPIOSign({
            'output_work_path': Artifact(Path),
            'task_names': List[str],
            'njobs': int,
            'task_paths': Artifact(List[Path])
        })

    @OP.exec_sign_check
    def execute(
            self,
            op_in: OPIO,
    ) -> OPIO:
        from ..core.common_prop import make_property_instance
        from ..core.calculator.calculator import make_calculator

        input_work_path = op_in["input_work_path"]
        path_to_prop = op_in["path_to_prop"]
        prop_param = op_in["prop_param"]
        inter_param = op_in["inter_param"]
        do_refine = op_in["do_refine"]

        cwd = Path.cwd()
        os.chdir(input_work_path)
        abs_path_to_prop = input_work_path / path_to_prop
        if os.path.exists(abs_path_to_prop):
            shutil.rmtree(abs_path_to_prop)
        create_path(str(abs_path_to_prop))
        conf_path = abs_path_to_prop.parent
        prop_name = abs_path_to_prop.name

        # break subworkflow if mismatch stop is set
        path_to_equi = conf_path / "relaxation" / "relax_task"
        try:
            structure_dict = loadfn(os.path.join(path_to_equi, "structure.json"))
        except FileNotFoundError:
            structure_dict = {}
        mismatch = structure_dict.get("mismatch", False)
        skip_mismatch = prop_param.get("skip_mismatch", False)
        if mismatch and skip_mismatch:
            print("Skipped due to mismatched relaxed structure")
            return OPIO({
                'output_work_path': abs_path_to_prop,
                'task_names': [],
                'njobs': 0,
                'task_paths': []
            })

        inter_param_prop = inter_param
        if "cal_setting" in prop_param and "overwrite_interaction" in prop_param["cal_setting"]:
            inter_param_prop = prop_param["cal_setting"]["overwrite_interaction"]

        prop = make_property_instance(prop_param, inter_param_prop)
        task_list = prop.make_confs(abs_path_to_prop, path_to_equi, do_refine)
        for kk in task_list:
            poscar = os.path.join(kk, "POSCAR")
            inter = make_calculator(inter_param_prop, poscar)
            inter.make_potential_files(kk)
            logging.debug(prop.task_type())  ### debug
            inter.make_input_file(kk, prop.task_type(), prop.task_param())
        prop.post_process(
            task_list
        )  # generate same KPOINTS file for elastic when doing VASP

        task_list.sort()
        os.chdir(path_to_prop)
        task_list_name = {'task_list': glob.glob('task.*').sort()}
        dumpfn(task_list_name, 'task_list.json')
        os.chdir(input_work_path)
        task_list_str = glob.glob(path_to_prop + '/' + 'task.*')
        task_list_str.sort()

        all_jobs = task_list
        njobs = len(all_jobs)
        jobs = [pathlib.Path(job) for job in task_list]

        os.chdir(cwd)
        op_out = OPIO({
            "output_work_path": input_work_path,
            "task_names": task_list_str,
            "njobs": njobs,
            "task_paths": jobs
        })
        return op_out


class PropsPost(OP):
    """
    OP class for analyzing calculation results (post property)
    """

    def __init__(self):
        pass

    @classmethod
    def get_input_sign(cls):
        return OPIOSign({
            'input_post': Artifact(Path, sub_path=False),
            'input_all': Artifact(Path),
            'prop_param': dict,
            'inter_param': dict,
            'task_names': List[str],
            'path_to_prop': str
        })

    @classmethod
    def get_output_sign(cls):
        return OPIOSign({
            'retrieve_path': Artifact(List[Path])
        })

    @OP.exec_sign_check
    def execute(self, op_in: OPIO) -> OPIO:
        from ..core.common_prop import make_property_instance
        cwd = os.getcwd()
        input_post = op_in["input_post"]
        input_all = op_in["input_all"]
        prop_param = op_in["prop_param"]
        inter_param = op_in["inter_param"]
        task_names = op_in["task_names"]
        path_to_prop = op_in["path_to_prop"]
        inter_type = inter_param["type"]

        if len(task_names) == 0:
            print("Skip post property")
            return OPIO({
                'retrieve_path': []
            })

        copy_dir_list_input = [path_to_prop.split('/')[0]]
        os.chdir(input_all)
        copy_dir_list = []
        for ii in copy_dir_list_input:
            copy_dir_list.extend(glob.glob(ii))
        copy_dir_list = list(set(copy_dir_list))
        copy_dir_list.sort()

        # find path of finished tasks
        os.chdir(op_in['input_post'])
        src_path = recursive_search(copy_dir_list)
        if not src_path:
            raise RuntimeError(f'Fail to find input work path after slices!')

        if inter_type in ['vasp', 'abacus']:
            os.chdir(input_post)
            for ii in task_names:
                shutil.copytree(os.path.join(ii, "backward_dir"), ii, dirs_exist_ok=True)
                shutil.rmtree(os.path.join(ii, "backward_dir"))
            os.chdir(input_all)
            shutil.copytree(input_post, './', dirs_exist_ok=True)
        else:
            os.chdir(input_all)
            # src_path = str(input_post) + str(local_path)
            shutil.copytree(src_path, './', dirs_exist_ok=True)

        if ("cal_setting" in prop_param
                and "overwrite_interaction" in prop_param["cal_setting"]):
            inter_param = prop_param["cal_setting"]["overwrite_interaction"]

        abs_path_to_prop = Path.cwd() / path_to_prop
        prop = make_property_instance(prop_param, inter_param)
        param_json = os.path.join(abs_path_to_prop, "param.json")
        param_dict = prop.parameter
        param_dict.setdefault("skip", False) # default of "skip" is False
        param_dict.pop("skip")
        dumpfn(param_dict, param_json)
        prop.compute(
            os.path.join(abs_path_to_prop, "result.json"),
            os.path.join(abs_path_to_prop, "result.out"),
            abs_path_to_prop,
        )
        # remove potential files in each task
        if inter_type in LAMMPS_INTER_TYPE:
            os.chdir(abs_path_to_prop)
            inter_files_name = []
            if type(inter_param["model"]) is str:
                inter_files_name = [inter_param["model"]]
            elif type(inter_param["model"]) is list:
                inter_files_name.extend(inter_param["model"])
            for file in inter_files_name:
                cmd = f"rm -f ../{file}"
                subprocess.call(cmd, shell=True)
                cmd = f"for kk in task.*; do rm -f $kk/{file}; done"
                subprocess.call(cmd, shell=True)
        elif inter_type == 'vasp':
            os.chdir(abs_path_to_prop)
            cmd = "rm -f ../POTCAR"
            subprocess.call(cmd, shell=True)
            cmd = f"for kk in task.*; do rm -f $kk/POTCAR; done"
            subprocess.call(cmd, shell=True)

        os.chdir(cwd)
        for ii in copy_dir_list:
            shutil.copytree(input_all / ii, ii, dirs_exist_ok=True)
        retrieve_path = [Path(ii) for ii in copy_dir_list]
        # out_path = Path(cwd) / 'retrieve_pool'
        # os.mkdir(out_path)
        # shutil.copytree(input_all / path_to_prop,
        #                out_path / path_to_prop, dirs_exist_ok=True)

        op_out = OPIO({
            'retrieve_path': retrieve_path
        })
        return op_out



from abc import ABC, abstractmethod


class StorageBase(ABC):
    def __init__(self, name):
        """
        Initiation of database

        Parameters
        ----------
        parameter : name
        """
        self.name = name

    @abstractmethod
    def sync(self, data: dict, id_field: str, depth: int):
        pass

    @abstractmethod
    def record(self, data: dict, id_field: str):
        pass

    @abstractmethod
    def load_json(self, query):
        pass

    @abstractmethod
    def close(self):
        pass



import argparse
import logging
import os
import datetime
from typing import List

from dflow import (
    Workflow,
    query_workflows,
    download_artifact
)

from apex import (
    header,
    __version__,
)
from apex.config import Config
from apex.step import do_step_from_args
from apex.submit import submit_from_args
from apex.archive import archive_from_args
from apex.report import report_from_args
from apex.utils import load_config_file


def parse_args():
    parser = argparse.ArgumentParser(
        description=f"APEX: A scientific workflow for Alloy Properties EXplorer "
                    f"using simulations (v{__version__})\n"
                    f"Type 'apex -h' for help.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    subparsers = parser.add_subparsers(title="Valid subcommands", dest="cmd")
    # version
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"APEX v{__version__}"
    )

    ##########################################
    # Submit
    parser_submit = subparsers.add_parser(
        "submit",
        help="Submit an APEX workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_submit.add_argument(
        "parameter", type=str, nargs='+',
        help='Json files to indicate calculation parameters'
    )
    parser_submit.add_argument(
        "-c", "--config",
        type=str, nargs='?',
        default='./global.json',
        help="The json file to config workflow",
    )
    parser_submit.add_argument(
        "-w", "--work",
        type=str, nargs='+',
        default='.',
        help="(Optional) Work directories to be submitted",
    )
    parser_submit.add_argument(
        "-d", "--debug",
        action="store_true",
        help="(Optional) Run APEX workflow via local debug mode"
    )
    parser_submit.add_argument(
        "-s", "--submit_only",
        action="store_true",
        help="(Optional) Submit workflow only without automatic result retrieving"
    )
    parser_submit.add_argument(
        '-f', "--flow",
        choices=['relax', 'props', 'joint'],
        help="(Optional) Specify type of workflow to submit: (relax | props | joint)"
    )
    parser_submit.add_argument(
        "-n", "--name",
        type=str, default=None,
        help="(Optional) Specify name of the workflow",
    )

    ##########################################
    # Do single step locally
    parser_do = subparsers.add_parser(
        "do",
        help="Run single step locally independent from workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_do.add_argument(
        "parameter", type=str,
        help='Json file to indicate calculation parameters'
    )
    parser_do.add_argument(
        "step",
        type=str,
        choices=[
            'make_relax', 'run_relax', 'post_relax',
            'make_props', 'run_props', 'post_props'
        ],
        help="Specify step name to be tested: "
             "(make_relax | run_relax | post_relax |"
             " make_props | run_props | post_props)"
    )
    parser_do.add_argument(
        "-c", "--config",
        type=str, nargs='?',
        default='./global.json',
        help="The json file to config the dpdispatcher",
    )

    ##########################################
    # Retrieve artifacts manually
    parser_retrieve = subparsers.add_parser(
        "retrieve",
        help="Retrieve results of an workflow with key provided manually",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_retrieve.add_argument(
        "-i", "--id", type=str, default=None,
        help='Workflow ID to be retrieved'
    )
    parser_retrieve.add_argument(
        "-w", "--work", type=str, default='.',
        help='Target work directory to be retrieved'
    )
    parser_retrieve.add_argument(
        "-c", "--config",
        type=str, nargs='?',
        default='./global.json',
        help="The json file to config workflow",
    )

    ##########################################
    ### dflow operations
    # list workflows
    parser_list = subparsers.add_parser(
        "list",
        help="List workflows",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_list.add_argument(
        "-l",
        "--label",
        type=str,
        default=None,
        help="query by labels",
    )
    parser_list.add_argument(
        "-c", "--config",
        type=str, nargs='?',
        default='./global.json',
        help="The json file to config workflow",
    )
    # get workflow
    parser_get = subparsers.add_parser(
        "get",
        help="Get a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_get.add_argument(
        "-i", "--id", type=str, default=None,
        help='Workflow ID to get'
    )
    parser_get.add_argument(
        "-w", "--work", type=str, default='.',
        help='Target work directory to get'
    )
    parser_get.add_argument(
        "-c", "--config",
        type=str, nargs='?',
        default='./global.json',
        help="The json file to config workflow",
    )
    # get steps of workflow
    parser_getsteps = subparsers.add_parser(
        "getsteps",
        help="Get steps from a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_getsteps.add_argument("ID", help="the workflow ID.")
    parser_getsteps.add_argument(
        "-n",
        "--name",
        type=str,
        default=None,
        help="query by name",
    )
    parser_getsteps.add_argument(
        "-k",
        "--key",
        type=str,
        default=None,
        help="query by key",
    )
    parser_getsteps.add_argument(
        "-p",
        "--phase",
        type=str,
        default=None,
        help="query by phase",
    )
    parser_getsteps.add_argument(
        "-i",
        "--id",
        type=str,
        default=None,
        help="workflow ID to query",
    )
    parser_getsteps.add_argument(
        "-w", "--work", type=str, default='.',
        help='Target work directory'
    )
    parser_getsteps.add_argument(
        "-t",
        "--type",
        type=str,
        default=None,
        help="query by type",
    )
    parser_getsteps.add_argument(
        "-c", "--config",
        type=str, nargs='?',
        default='./global.json',
        help="The json file to config workflow",
    )
    #  getkeys of workflow
    parser_getkeys = subparsers.add_parser(
        "getkeys",
        help="Get keys of steps from a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_getkeys.add_argument(
        "-i", "--id", type=str, default=None,
        help='Workflow ID to get keys'
    )
    parser_getkeys.add_argument(
        "-w", "--work", type=str, default='.',
        help='Target work directory get keys'
    )
    parser_getkeys.add_argument(
        "-c", "--config",
        type=str, nargs='?',
        default='./global.json',
        help="The json file to config workflow",
    )
    # delete workflow
    parser_delete = subparsers.add_parser(
        "delete",
        help="Delete a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_delete.add_argument(
        "-i", "--id", type=str, default=None,
        help='Workflow ID to delete'
    )
    parser_delete.add_argument(
        "-w", "--work", type=str, default='.',
        help='Target work directory delete'
    )
    parser_delete.add_argument(
        "-c", "--config",
        type=str, nargs='?',
        default='./global.json',
        help="The json file to config workflow",
    )
    # resubmit workflow
    parser_resubmit = subparsers.add_parser(
        "resubmit",
        help="Resubmit a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_resubmit.add_argument(
        "-i", "--id", type=str, default=None,
        help='Workflow ID to resubmit'
    )
    parser_resubmit.add_argument(
        "-w", "--work", type=str, default='.',
        help='Target work directory to resubmit'
    )
    parser_resubmit.add_argument(
        "-c", "--config",
        type=str, nargs='?',
        default='./global.json',
        help="The json file to config workflow",
    )
    # retry workflow
    parser_retry = subparsers.add_parser(
        "retry",
        help="Retry a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_retry.add_argument(
        "-i", "--id", type=str, default=None,
        help='Workflow ID to retry'
    )
    parser_retry.add_argument(
        "-w", "--work", type=str, default='.',
        help='Target work directory to retry'
    )
    parser_retry.add_argument(
        "-s",
        "--step",
        type=str,
        default=None,
        help="retry a step in a running workflow with step ID (experimental)",
    )
    # resume workflow
    parser_resume = subparsers.add_parser(
        "resume",
        help="Resume a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_resume.add_argument(
        "-i", "--id", type=str, default=None,
        help='Workflow ID to resume'
    )
    parser_resume.add_argument(
        "-w", "--work", type=str, default='.',
        help='Target work directory to resume'
    )
    parser_resume.add_argument(
        "-c", "--config",
        type=str, nargs='?',
        default='./global.json',
        help="The json file to config workflow",
    )
    # stop workflow
    parser_stop = subparsers.add_parser(
        "stop",
        help="Stop a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_stop.add_argument(
        "-i", "--id", type=str, default=None,
        help='Workflow ID to stop'
    )
    parser_stop.add_argument(
        "-w", "--work", type=str, default='.',
        help='Target work directory to stop'
    )
    parser_stop.add_argument(
        "-c", "--config",
        type=str, nargs='?',
        default='./global.json',
        help="The json file to config workflow",
    )
    # suspend workflow
    parser_suspend = subparsers.add_parser(
        "suspend",
        help="Suspend a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_suspend.add_argument(
        "-i", "--id", type=str, default=None,
        help='Workflow ID to suspend'
    )
    parser_suspend.add_argument(
        "-w", "--work", type=str, default='.',
        help='Target work directory to suspend'
    )
    parser_suspend.add_argument(
        "-c", "--config",
        type=str, nargs='?',
        default='./global.json',
        help="The json file to config workflow",
    )
    # terminate workflow
    parser_terminate = subparsers.add_parser(
        "terminate",
        help="Terminate a workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_terminate.add_argument(
        "-i", "--id", type=str, default=None,
        help='Workflow ID to terminate'
    )
    parser_terminate.add_argument(
        "-w", "--work", type=str, default='.',
        help='Target work directory to terminate'
    )
    parser_terminate.add_argument(
        "-c", "--config",
        type=str, nargs='?',
        default='./global.json',
        help="The json file to config workflow",
    )

    ##########################################
    # Archive results
    parser_archive = subparsers.add_parser(
        "archive",
        help="Archive test results to local or database",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_archive.add_argument(
        "json", type=str, nargs='+',
        help='Json files to indicate calculation parameters '
             'or result json files that will be directly archived to database when -r flag is raised'
    )
    parser_archive.add_argument(
        "-c", "--config",
        type=str, nargs='?',
        default='./global.json',
        help="The json file of global config",
    )
    parser_archive.add_argument(
        "-w", "--work",
        type=str, nargs='+',
        default='.',
        help="(Optional) Working directory",
    )
    parser_archive.add_argument(
        '-f', "--flow",
        choices=['relax', 'props', 'joint'],
        help="(Optional) Specify type of workflow's results to archive: (relax | props | joint)"
    )
    parser_archive.add_argument(
        '-d', "--database",
        choices=['local', 'mongodb', 'dynamodb'],
        help="(Optional) Specify type of database: (local | mongodb | dynamodb)"
    )
    parser_archive.add_argument(
        '-m', "--method",
        choices=['sync', 'record'],
        help="(Optional) Specify archive method: (sync | record)"
    )
    parser_archive.add_argument(
        "-t", "--tasks",
        action="store_true",
        help="Whether to archive running details of each task (default: False)"
    )
    parser_archive.add_argument(
        "-r", "--result",
        action="store_true",
        help="(Optional) whether to treat json files as results and archive them directly to database",
    )

    ##########################################
    # Report results
    parser_report = subparsers.add_parser(
        "report",
        help="Run result visualization report app",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_report.add_argument(
        "-c", "--config",
        type=str, nargs='?',
        default='./global.json',
        help="The json file of global config",
    )
    parser_report.add_argument(
        "-w", "--work",
        type=str, nargs='+',
        default='.',
        help="(Optional) Working directory or json file path to be reported",
    )

    parsed_args = parser.parse_args()
    # print help if no parser
    if not parsed_args.cmd:
        parser.print_help()

    return parser, parsed_args


def config_dflow(config_file: os.PathLike) -> None:
    # config dflow_config and s3_config
    config_dict = load_config_file(config_file)
    wf_config = Config(**config_dict)
    Config.config_dflow(wf_config.dflow_config_dict)
    Config.config_bohrium(wf_config.bohrium_config_dict)
    Config.config_s3(wf_config.dflow_s3_config_dict)


def format_print_table(t: List[List[str]]):
    ncol = len(t[0])
    maxlen = [0] * ncol
    for row in t:
        for i, s in enumerate(row):
            if len(str(s)) > maxlen[i]:
                maxlen[i] = len(str(s))
    for row in t:
        for i, s in enumerate(row):
            print(str(s) + " " * (maxlen[i]-len(str(s))+3), end="")
        print()


def format_time_delta(td: datetime.timedelta) -> str:
    if td.days > 0:
        return "%dd" % td.days
    elif td.seconds >= 3600:
        return "%dh" % (td.seconds // 3600)
    else:
        return "%ds" % td.seconds


def get_id_from_record(work_dir: os.PathLike, operation_name: str = None) -> str:
    logging.info(msg='No workflow_id is provided, will employ the latest workflow')
    workflow_log = os.path.join(work_dir, '.workflow.log')
    assert os.path.isfile(workflow_log), \
        'No workflow_id is provided and no .workflow.log file found in work_dir'
    with open(workflow_log, 'r') as f:
        try:
            last_record = f.readlines()[-1]
        except IndexError:
            raise RuntimeError('No workflow_id is provided and .workflow.log file is empty!')
    workflow_id = last_record.split('\t')[0]
    assert workflow_id, 'No workflow ID for operation!'
    logging.info(msg=f'Operating on workflow ID: {workflow_id}')
    if operation_name:
        modified_record = last_record.split('\t')
        modified_record[1] = operation_name
        modified_record[2] = datetime.datetime.now().isoformat()
        with open(workflow_log, 'a') as f:
            f.write('\t'.join(modified_record))
    return workflow_id


def main():
    # logging
    logging.basicConfig(level=logging.INFO)
    # parse args
    parser, args = parse_args()
    if args.cmd == 'submit':
        header()
        submit_from_args(
            parameters=args.parameter,
            config_file=args.config,
            work_dirs=args.work,
            indicated_flow_type=args.flow,
            flow_name=args.name,
            submit_only=args.submit_only,
            is_debug=args.debug
        )
    elif args.cmd == "list":
            config_dflow(args.config)
            if args.label is not None:
                labels = {}
                for label in args.label.split(","):
                    key, value = label.split("=")
                    labels[key] = value
            else:
                labels = None
            wfs = query_workflows(labels=labels)
            t = [["NAME", "STATUS", "AGE", "DURATION"]]
            for wf in wfs:
                tc = datetime.datetime.strptime(wf.metadata.creationTimestamp,
                                                "%Y-%m-%dT%H:%M:%SZ")
                age = format_time_delta(datetime.datetime.now() - tc)
                dur = format_time_delta(wf.get_duration())
                t.append([wf.id, wf.status.phase, age, dur])
            format_print_table(t)
    elif args.cmd == "get":
        config_dflow(args.config)
        wf_id = args.id
        if not wf_id:
            wf_id = get_id_from_record(args.work, 'get')
        wf = Workflow(id=wf_id)
        info = wf.query()
        t = []
        t.append(["Name:", info.id])
        t.append(["Status:", info.status.phase])
        t.append(["Created:", info.metadata.creationTimestamp])
        t.append(["Started:", info.status.startedAt])
        t.append(["Finished:", info.status.finishedAt])
        t.append(["Duration", format_time_delta(info.get_duration())])
        t.append(["Progress:", info.status.progress])
        format_print_table(t)
        print()
        steps = info.get_step()
        t = [["STEP", "ID", "KEY", "TYPE", "PHASE", "DURATION"]]
        for step in steps:
            if step.type in ["StepGroup"]:
                continue
            key = step.key if step.key is not None else ""
            dur = format_time_delta(step.get_duration())
            t.append([step.displayName, step.id, key, step.type, step.phase,
                      dur])
        format_print_table(t)
    elif args.cmd == "getsteps":
        config_dflow(args.config)
        wf_id = args.ID
        name = args.name
        key = args.key
        phase = args.phase
        id = args.id
        type = args.type
        if name is not None:
            name = name.split(",")
        if key is not None:
            key = key.split(",")
        if phase is not None:
            phase = phase.split(",")
        if id is not None:
            id = id.split(",")
        if type is not None:
            type = type.split(",")
        wf = Workflow(id=wf_id)
        if key is not None:
            steps = wf.query_step_by_key(key, name, phase, id, type)
        else:
            steps = wf.query_step(name, key, phase, id, type)
        for step in steps:
            if step.type in ["StepGroup"]:
                continue
            key = step.key if step.key is not None else ""
            dur = format_time_delta(step.get_duration())
            t = []
            t.append(["Step:", step.displayName])
            t.append(["ID:", step.id])
            t.append(["Key:", key])
            t.append(["Type:", step.type])
            t.append(["Phase:", step.phase])
            format_print_table(t)
            if hasattr(step, "outputs"):
                if hasattr(step.outputs, "parameters"):
                    print("Output parameters:")
                    for name, par in step.outputs.parameters.items():
                        if name[:6] == "dflow_":
                            continue
                        print("%s: %s" % (name, par.value))
                    print()
                if hasattr(step.outputs, "artifacts"):
                    print("Output artifacts:")
                    for name, art in step.outputs.artifacts.items():
                        if name[:6] == "dflow_" or name == "main-logs":
                            continue
                        key = ""
                        if hasattr(art, "s3"):
                            key = art.s3.key
                        elif hasattr(art, "oss"):
                            key = art.oss.key
                        print("%s: %s" % (name, key))
                    print()
            print()
    elif args.cmd == "getkeys":
        config_dflow(args.config)
        wf_id = args.id
        if not wf_id:
            wf_id = get_id_from_record(args.work, 'getkeys')
        wf = Workflow(id=wf_id)
        keys = wf.query_keys_of_steps()
        print("\n".join(keys))
    elif args.cmd == "delete":
        config_dflow(args.config)
        wf_id = args.id
        if not wf_id:
            wf_id = get_id_from_record(args.work, 'delete')
        wf = Workflow(id=wf_id)
        wf.delete()
        print(f'Workflow deleted! (ID: {wf.id}, UID: {wf.uid})')
    elif args.cmd == "resubmit":
        config_dflow(args.config)
        wf_id = args.id
        if not wf_id:
            wf_id = get_id_from_record(args.work, 'resubmit')
        wf = Workflow(id=wf_id)
        wf.resubmit()
        print(f'Workflow resubmitted... (ID: {wf.id}, UID: {wf.uid})')
    elif args.cmd == "resume":
        config_dflow(args.config)
        wf_id = args.id
        if not wf_id:
            wf_id = get_id_from_record(args.work, 'resume')
        wf = Workflow(id=wf_id)
        wf.resume()
        print(f'Workflow resumed... (ID: {wf.id}, UID: {wf.uid})')
    elif args.cmd == "retry":
        config_dflow(args.config)
        wf_id = args.id
        if not wf_id:
            wf_id = get_id_from_record(args.work, 'retry')
        wf = Workflow(id=wf_id)
        if args.step is not None:
            wf.retry_steps(args.step.split(","))
        else:
            wf.retry()
        print(f'Workflow retried... (ID: {wf.id}, UID: {wf.uid})')
    elif args.cmd == "stop":
        config_dflow(args.config)
        wf_id = args.id
        if not wf_id:
            wf_id = get_id_from_record(args.work, 'stop')
        wf = Workflow(id=wf_id)
        wf.stop()
        print(f'Workflow stopped! (ID: {wf.id}, UID: {wf.uid})')
    elif args.cmd == "suspend":
        config_dflow(args.config)
        wf_id = args.id
        if not wf_id:
            wf_id = get_id_from_record(args.work, 'suspend')
        wf = Workflow(id=wf_id)
        wf.suspend()
        print(f'Workflow suspended... (ID: {wf.id}, UID: {wf.uid})')
    elif args.cmd == "terminate":
        config_dflow(args.config)
        wf_id = args.id
        if not wf_id:
            wf_id = get_id_from_record(args.work, 'terminate')
        wf = Workflow(id=wf_id)
        wf.terminate()
    elif args.cmd == 'retrieve':
        config_dflow(args.config)
        wf_id = args.id
        if not wf_id:
            wf_id = get_id_from_record(args.work, 'retrieve')
        wf = Workflow(id=wf_id)
        work_dir = args.work
        all_keys = wf.query_keys_of_steps()
        wf_info = wf.query()
        download_keys = [key for key in all_keys if key.split('-')[0] == 'propertycal' or key == 'relaxationcal']
        task_left = len(download_keys)
        print(f'Retrieving {task_left} workflow results {wf_id} to {work_dir}')

        for key in download_keys:
            step = wf_info.get_step(key=key)[0]
            task_left -= 1
            if step['phase'] == 'Succeeded':
                logging.info(f"Retrieving {key}...({task_left} more left)")
                download_artifact(
                    artifact=step.outputs.artifacts['retrieve_path'],
                    path=work_dir
                )
            else:
                logging.warning(f"Step {key} with status: {step['phase']} will be skipping...({task_left} more left)")
    elif args.cmd == 'do':
        header()
        do_step_from_args(
            parameter=args.parameter,
            machine_file=args.config,
            step=args.step
        )
    elif args.cmd == 'archive':
        archive_from_args(
            parameters=args.json,
            config_file=args.config,
            work_dirs=args.work,
            indicated_flow_type=args.flow,
            database_type=args.database,
            method=args.method,
            archive_tasks=args.tasks,
            is_result=args.result
        )
    elif args.cmd == 'report':
        header()
        report_from_args(
            config_file=args.config,
            path_list=args.work,
        )
    else:
        raise RuntimeError(
            f"unknown command {args.cmd}\n{parser.print_help()}"
        )



import glob
import logging
import json
import os
from typing import List
from pathlib import Path
from monty.json import MontyEncoder
from monty.serialization import loadfn, dumpfn
from apex.utils import (
    judge_flow,
    json2dict,
    update_dict,
    return_prop_list,
    load_config_file,
    generate_random_string
)
from apex.database.DatabaseFactory import DatabaseFactory
from apex.config import Config


class ResultStorage:
    def __init__(self, work_dir):
        self.work_dir = Path(work_dir).absolute()
        self.result_dict = {'work_path': str(self.work_dir)}

    @property
    def result_data(self):
        return self.result_dict

    @property
    def work_dir_path(self):
        return str(self.work_dir)

    @json2dict
    def sync_relax(self, relax_param: dict):
        # sync results from relaxation task
        confs = relax_param["structures"]
        interaction = relax_param["interaction"]
        conf_dirs = []
        for conf in confs:
            conf_dirs.extend(glob.glob(str(self.work_dir / conf)))
        conf_dirs.sort()
        for ii in conf_dirs:
            relax_task = os.path.join(ii, 'relaxation/relax_task')
            inter,task,structure,result=[os.path.join(relax_task,ii) for ii in 
                ['inter.json','task.json','structure.json','result.json']]
            if not (
                os.path.isfile(inter)
                and os.path.isfile(task)
                and os.path.isfile(result)
            ):
                logging.warning(
                    f"relaxation result path is not complete, will skip result extraction from {relax_task}"
                )
                continue
            logging.info(msg=f"extract results from {relax_task}")
            conf_key = os.path.relpath(ii, self.work_dir)
            conf_dict = {"interaction": loadfn(inter),
                         "parameter": loadfn(task),
                         "structure_info": loadfn(structure),
                         "result": loadfn(result)}
            new_dict = {conf_key: {"relaxation": conf_dict}}
            update_dict(self.result_dict, new_dict)

    @json2dict
    def sync_props(self, props_param: dict, archive_tasks: bool = False):
        # sync results from property test
        confs = props_param["structures"]
        interaction = props_param["interaction"]
        properties = props_param["properties"]
        prop_list = return_prop_list(properties)
        conf_dirs = []
        for conf in confs:
            conf_dirs.extend(glob.glob(str(self.work_dir / conf)))
        conf_dirs.sort()
        for ii in conf_dirs:
            for jj in prop_list:
                prop_dir = os.path.join(ii, jj)
                result = os.path.join(prop_dir, 'result.json')
                param = os.path.join(prop_dir, 'param.json')
                task_list_path = os.path.join(prop_dir, 'task_list.json')
                if not os.path.isfile(result):
                    logging.warning(
                        f"Property post-process is not complete, will skip result extraction from {prop_dir}"
                    )
                    continue
                logging.info(msg=f"extract results from {prop_dir}")
                conf_key = os.path.relpath(ii, self.work_dir)
                result_dict = loadfn(result)
                try:
                    param_dict = loadfn(param)
                except FileNotFoundError:
                    logging.warning(f'{param} file not found')
                    param_dict = None
                prop_dict = {"parameter": param_dict, "result": result_dict}
                # extract running details of each task
                if archive_tasks:
                    logging.debug(msg='Archive running details of tasks...')
                    logging.warning(
                        msg='You are trying to archive detailed running log of each task into database,'
                            'which may exceed the limitation of database allowance.'
                            'Please consider spliting the data or only archiving details of the most important property.'
                    )
                    try:
                        task_list = loadfn(task_list_path)
                        result_task_path = [os.path.join(prop_dir, task, 'result_task.json') for task in task_list]
                    except FileNotFoundError:
                        logging.warning(f'{task_list_path} file not found, will track all tasks listed {prop_dir}')
                        result_task_path = glob.glob(os.path.join(prop_dir, 'task.*', 'result_task.json'))
                    task_result_list = [loadfn(kk) for kk in sorted(result_task_path)]
                    prop_dict["tasks"] = task_result_list

                new_dict = {conf_key: {jj: prop_dict}}
                update_dict(self.result_dict, new_dict)


def connect_database(config):
    # connect to database
    if config.database_type == 'mongodb':
        logging.info(msg='Use database type: MongoDB')
        database = DatabaseFactory.create_database(
            'mongodb',
            'mongodb',
            config.mongodb_database,
            config.mongodb_collection,
            **config.mongodb_config_dict
        )
    elif config.database_type == 'dynamodb':
        logging.info(msg='Use database type: DynamoDB')
        database = DatabaseFactory.create_database(
            'dynamodb',
            'dynamodb',
            config.dynamodb_table_name,
            **config.dynamodb_config_dict
        )
    else:
        raise RuntimeError(f'unsupported database type: {config.database_type}')
    return database


def archive2db(config, data: dict, data_id: str):
    database = connect_database(config)
    # archive results database
    if config.archive_method == 'sync':
        logging.debug(msg='Archive method: sync')
        database.sync(data, data_id, depth=2)
    elif config.archive_method == 'record':
        logging.debug(msg='Archive method: record')
        database.record(data, data_id)
    else:
        raise TypeError(
            f'Unrecognized archive method: {config.archive_method}. '
            f"Should select from 'sync' and 'record'."
        )


def archive_workdir(relax_param, props_param, config, work_dir, flow_type):
    print(f'=> Begin archiving {work_dir}')
    # extract results json
    store = ResultStorage(work_dir)
    if relax_param and flow_type != 'props':
        store.sync_relax(relax_param)
    if props_param and flow_type != 'relax':
        store.sync_props(props_param, config.archive_tasks)

    dump_file = os.path.join(store.work_dir_path, 'all_result.json')
    default_id = generate_random_string(10)
    if os.path.isfile(dump_file):
        logging.info(msg='all_result.json exists, and will be updated.')
        orig_data = loadfn(dump_file)
        try:
            default_id = orig_data['archive_key']
        except KeyError:
            store.result_data['archive_key'] = default_id
        update_dict(orig_data, store.result_data, depth=2)
        dumpfn(orig_data, dump_file, indent=4)
    else:
        store.result_data['archive_key'] = default_id
        dumpfn(store.result_data, dump_file, indent=4)

    # try to get documented key id from all_result.json
    # define archive key
    data_id = config.archive_key if config.archive_key else default_id

    if config.database_type != 'local':
        data_json_str = json.dumps(store.result_data, cls=MontyEncoder, indent=4)
        data_dict = json.loads(data_json_str)
        data_dict['_id'] = data_id

        archive2db(config, data_dict, data_id)


def archive2db_from_json(config, json_file):
    logging.info(msg=f'Archive from local json file: {json_file}')
    data_dict = loadfn(json_file)
    data_json_str = json.dumps(data_dict, cls=MontyEncoder, indent=4)
    data_dict = json.loads(data_json_str)
    # define archive key
    if config.archive_key:
        data_id = config.archive_key
    else:
        data_id = data_dict['archive_key']
    data_dict['_id'] = data_id

    archive2db(config, data_dict, data_id)


def archive_result(
        parameters: List[os.PathLike],
        config_dict: dict,
        work_dir: List[os.PathLike],
        indicated_flow_type: str,
        database_type,
        method,
        archive_tasks,
        is_result
):
    global_config = Config(**config_dict)
    # re-specify args
    if database_type:
        global_config.database_type = database_type
    if method:
        global_config.archive_method = method
    if archive_tasks:
        global_config.archive_tasks = archive_tasks

    if is_result:
        # archive local results json file
        json_file_list = []
        # Parameter here stands for json files that store test results and be archived directly
        for ii in parameters:
            glob_list = glob.glob(os.path.abspath(ii))
            json_file_list.extend(glob_list)
            json_file_list.sort()
        for ii in json_file_list:
            archive2db_from_json(global_config, ii)
    else:
        _, _, flow_type, relax_param, props_param = judge_flow(
            [loadfn(jj) for jj in parameters],
            indicated_flow_type
        )
        # archive work directories
        work_dir_list = []
        for ii in work_dir:
            glob_list = glob.glob(os.path.abspath(ii))
            work_dir_list.extend(glob_list)
            work_dir_list.sort()
        for ii in work_dir_list:
            archive_workdir(relax_param, props_param, global_config, ii, flow_type)


def archive_from_args(
        parameters: List[os.PathLike],
        config_file: os.PathLike,
        work_dirs: List[os.PathLike],
        indicated_flow_type: str,
        database_type,
        method,
        archive_tasks,
        is_result
):
    print('-------Archive result Mode-------')
    archive_result(
        parameters=parameters,
        config_dict=load_config_file(config_file),
        work_dir=work_dirs,
        indicated_flow_type=indicated_flow_type,
        database_type=database_type,
        method=method,
        archive_tasks=archive_tasks,
        is_result=is_result
    )
    print('Complete!')



import logging
import json
import numpy as np
import plotly.graph_objs as go
from dash import dash_table
import pandas as pd
from monty.json import MontyEncoder

from apex.core.lib.utils import round_format, round_2d_format

TABLE_WIDTH = '100%'
TABLE_MIN_WIDTH = '100%'


class RelaxationReport:
    @staticmethod
    def dash_table(res_data: dict, decimal: int = 3, **kwargs) -> dash_table.DataTable:
        conf_list = []
        equi_en = []
        cell_vec_a = []
        cell_vec_b = []
        cell_vec_c = []
        space_group_symbol = []
        space_group_number = []
        point_group_symbol = []
        crystal_system = []
        lattice_type = []
        for conf, dataset in res_data.items():
            try:
                class_data = dataset['relaxation']['result']
                struct_info = dataset['relaxation']['structure_info']
            except KeyError:
                pass
            else:
                data = json.dumps(class_data, cls=MontyEncoder, indent=4)
                data = json.loads(data)
                conf_list.append(conf)
                equi_en.append(data["data"]["energies"]["data"][-1])
                vec_a_length = np.linalg.norm(data["data"]["cells"]["data"][-1][0])
                vec_b_length = np.linalg.norm(data["data"]["cells"]["data"][-1][1])
                vec_c_length = np.linalg.norm(data["data"]["cells"]["data"][-1][2])
                cell_vec_a.append(vec_a_length)
                cell_vec_b.append(vec_b_length)
                cell_vec_c.append(vec_c_length)
                space_group_symbol.append(struct_info["space_group_symbol"])
                space_group_number.append(struct_info["space_group_number"])
                point_group_symbol.append(struct_info["point_group_symbol"])
                crystal_system.append(struct_info["crystal_system"])
                lattice_type.append(struct_info["lattice_type"])

        # round numbers in table
        # rounded_tensor = round_2d_format(data, decimal)

        df = pd.DataFrame({
            "Conf": conf_list,
            "Equi E (eV)": equi_en,
            "Cell Vector length a (Å)": cell_vec_a,
            "Cell Vector length b (Å)": cell_vec_b,
            "Cell Vector length c (Å)": cell_vec_c,
            "Space Group Symbol": space_group_symbol,
            "Space Group Number": space_group_number,
            "Point Group Symbol": point_group_symbol,
            "Crystal System": crystal_system,
            "Lattice Type": lattice_type,
        })

        table = dash_table.DataTable(
            data=df.to_dict('records'),
            columns=[{'name': i, 'id': i} for i in df.columns],
            style_table={'width': TABLE_WIDTH,
                         'minWidth': TABLE_MIN_WIDTH,
                         'overflowX': 'auto'},
            style_cell={'textAlign': 'left', 'width': '150px'}
        )

        return table



import os
import os.path
import glob
import shutil
import tempfile
import logging
from typing import List
from multiprocessing import Pool
from monty.serialization import loadfn

import apex
import dpdata
import fpop
from dflow import config, s3_config

from apex.archive import archive_workdir
from apex.config import Config
from apex.flow import FlowGenerator
from apex.utils import (
    judge_flow,
    load_config_file,
    json2dict,
    copy_all_other_files,
    sepline,
    handle_prop_suffix,
    backup_path
)


def pack_upload_dir(
        work_dir: os.PathLike,
        upload_dir: os.PathLike,
        relax_param: dict,
        prop_param: dict,
        flow_type: str,
        exclude_upload_files: List[str],
):
    """
    Pack the necessary files and directories into temp dir and upload it to dflow
    """
    cwd = os.getcwd()
    os.chdir(work_dir)
    relax_confs = relax_param.get("structures", []) if relax_param else []
    prop_confs = prop_param.get("structures", []) if prop_param else []
    relax_prefix = relax_param["interaction"].get("potcar_prefix", None) if relax_param else None
    prop_prefix = prop_param["interaction"].get("potcar_prefix", None) if prop_param else None
    include_dirs = set()
    if relax_prefix:
        relax_prefix_base = relax_prefix.split('/')[0]
        include_dirs.add(relax_prefix_base)
    if prop_prefix:
        prop_prefix_base = prop_prefix.split('/')[0]
        include_dirs.add(prop_prefix_base)
    confs = relax_confs + prop_confs
    assert len(confs) > 0, "No configuration path indicated!"
    conf_dirs = []
    for conf in confs:
        conf_dirs.extend(glob.glob(conf))
    conf_dirs = list(set(conf_dirs))
    conf_dirs.sort()
    refine_init_name_list = []
    # backup all existing property work directories
    if flow_type in ['props', 'joint']:
        property_list = prop_param["properties"]
        for ii in conf_dirs:
            sepline(ch=ii, screen=True)
            for jj in property_list:
                do_refine, suffix = handle_prop_suffix(jj)
                property_type = jj["type"]
                if not suffix:
                    continue
                if do_refine:
                    refine_init_suffix = jj['init_from_suffix']
                    refine_init_name_list.append(property_type + "_" + refine_init_suffix)
                path_to_prop = os.path.join(ii, property_type + "_" + suffix)
                backup_path(path_to_prop)

    """copy necessary files and directories into temp upload directory"""
    exclude_upload_files.append("all_result.json")
    copy_all_other_files(
        work_dir, upload_dir,
        exclude_files=exclude_upload_files,
        include_dirs=list(include_dirs)
    )
    for ii in conf_dirs:
        build_conf_path = os.path.join(upload_dir, ii)
        os.makedirs(build_conf_path, exist_ok=True)
        copy_poscar_path = os.path.abspath(os.path.join(ii, "POSCAR"))
        copy_stru_path = os.path.abspath(os.path.join(ii, "STRU"))
        if os.path.isfile(copy_poscar_path):
            target_poscar_path = os.path.join(build_conf_path, "POSCAR")
            shutil.copy(copy_poscar_path, target_poscar_path)
        if os.path.isfile(copy_stru_path):
            target_stru_path = os.path.join(build_conf_path, "STRU")
            shutil.copy(copy_stru_path, target_stru_path)
        if flow_type == 'props':
            copy_relaxation_path = os.path.abspath(os.path.join(ii, "relaxation"))
            target_relaxation_path = os.path.join(build_conf_path, "relaxation")
            shutil.copytree(copy_relaxation_path, target_relaxation_path)
            # copy refine from init path to upload dir
            if refine_init_name_list:
                for jj in refine_init_name_list:
                    copy_init_path = os.path.abspath(os.path.join(ii, jj))
                    assert os.path.exists(copy_init_path), f'refine from init path {copy_init_path} does not exist!'
                    target_init_path = os.path.join(build_conf_path, jj)
                    shutil.copytree(copy_init_path, target_init_path)

    os.chdir(cwd)


def submit(
        flow,
        flow_type,
        work_dir,
        relax_param,
        props_param,
        wf_config,
        conf=config,
        s3_conf=s3_config,
        is_sub=False,
        labels=None,
):
    if is_sub:
        # reset dflow global config for sub-processes
        logging.info(msg=f'Sub-process working on: {work_dir}')
        config.update(conf)
        s3_config.update(s3_conf)
        logging.basicConfig(level=logging.INFO)
    else:
        logging.info(msg=f'Working on: {work_dir}')

    with tempfile.TemporaryDirectory() as tmp_dir:
        logging.debug(msg=f'Temporary upload directory:{tmp_dir}')
        pack_upload_dir(
            work_dir=work_dir,
            upload_dir=tmp_dir,
            relax_param=relax_param,
            prop_param=props_param,
            flow_type=flow_type,
            exclude_upload_files=wf_config.exclude_upload_files
        )

        cwd = os.getcwd()
        os.chdir(tmp_dir)
        flow_id = None
        flow_name = wf_config.flow_name
        submit_only = wf_config.submit_only
        if flow_type == 'relax':
            flow_id = flow.submit_relax(
                upload_path=tmp_dir,
                download_path=work_dir,
                relax_parameter=relax_param,
                submit_only=submit_only,
                name=flow_name,
                labels=labels
            )
        elif flow_type == 'props':
            flow_id = flow.submit_props(
                upload_path=tmp_dir,
                download_path=work_dir,
                props_parameter=props_param,
                submit_only=submit_only,
                name=flow_name,
                labels=labels
            )
        elif flow_type == 'joint':
            flow_id = flow.submit_joint(
                upload_path=tmp_dir,
                download_path=work_dir,
                props_parameter=props_param,
                relax_parameter=relax_param,
                submit_only=submit_only,
                name=flow_name,
                labels=labels
            )
        os.chdir(cwd)

    if not submit_only:
        # auto archive results
        print(f'Archiving results of workflow (ID: {flow_id}) into {wf_config.database_type}...')
        archive_workdir(relax_param, props_param, wf_config, work_dir, flow_type)


def submit_workflow(
    parameter_dicts: List[dict],
    config_dict: dict,
    work_dirs: List[os.PathLike],
    indicated_flow_type: str,
    flow_name: str = None,
    submit_only=False,
    is_debug=False,
    labels=None
):
    # config dflow_config and s3_config
    wf_config = Config(**config_dict)
    Config.config_dflow(wf_config.dflow_config_dict)
    Config.config_bohrium(wf_config.bohrium_config_dict)
    Config.config_s3(wf_config.dflow_s3_config_dict)
    if submit_only:
        print('Submit only mode activated, no auto-retrieval of results.')
        wf_config.submit_only = True
    # set pre-defined dflow debug mode settings
    if is_debug:
        tmp_work_dir = tempfile.TemporaryDirectory()
        config["mode"] = "debug"
        config["debug_workdir"] = config_dict.get("debug_workdir", tmp_work_dir.name)
        logging.info(f'Debug mode activated, debug work directory: {config["debug_workdir"]}')
        s3_config["storage_client"] = None

    if flow_name:
        wf_config.flow_name = flow_name

    # judge basic flow info from user indicated parameter files
    (run_op, calculator, flow_type,
     relax_param, props_param) = judge_flow(parameter_dicts, indicated_flow_type)
    print(f'Running APEX calculation via {calculator}')
    print(f'Submitting {flow_type} workflow...')
    make_image = wf_config.basic_config_dict["apex_image_name"]
    run_image = wf_config.basic_config_dict[f"{calculator}_image_name"]
    if not run_image:
        run_image = wf_config.basic_config_dict["run_image_name"]
    run_command = wf_config.basic_config_dict[f"{calculator}_run_command"]
    if not run_command:
        run_command = wf_config.basic_config_dict["run_command"]
    post_image = make_image
    group_size = wf_config.basic_config_dict["group_size"]
    pool_size = wf_config.basic_config_dict["pool_size"]
    executor = wf_config.get_executor(wf_config.dispatcher_config_dict)

    # upload necessary python dependencies
    upload_python_packages = wf_config.basic_config_dict["upload_python_packages"]
    upload_python_packages.extend(list(apex.__path__))
    upload_python_packages.extend(list(fpop.__path__))
    upload_python_packages.extend(list(dpdata.__path__))
    #upload_python_packages.extend(list(phonolammps.__path__))

    flow = FlowGenerator(
        make_image=make_image,
        run_image=run_image,
        post_image=post_image,
        run_command=run_command,
        calculator=calculator,
        run_op=run_op,
        group_size=group_size,
        pool_size=pool_size,
        executor=executor,
        upload_python_packages=upload_python_packages
    )
    # submit the workflows
    work_dir_list = []
    for ii in work_dirs:
        glob_list = glob.glob(os.path.abspath(ii))
        work_dir_list.extend(glob_list)
        work_dir_list.sort()
    if len(work_dir_list) > 1:
        n_processes = len(work_dir_list)
        print(f'Submitting via {n_processes} processes...')
        pool = Pool(processes=n_processes)
        for ii in work_dir_list:
            res = pool.apply_async(
                submit,
                (flow,
                 flow_type,
                 ii,
                 relax_param,
                 props_param,
                 wf_config,
                 config,
                 s3_config,
                 True,
                 labels)
            )
        pool.close()
        pool.join()
    elif len(work_dir_list) == 1:
        submit(
            flow,
            flow_type,
            work_dir_list[0],
            relax_param,
            props_param,
            wf_config,
            labels=labels,
        )
    else:
        raise NotADirectoryError('Empty work directory indicated, please check your argument')


def submit_from_args(
        parameters,
        config_file: os.PathLike,
        work_dirs,
        indicated_flow_type: str,
        flow_name: str = None,
        submit_only=False,
        is_debug=False,
):
    print('-------Submit Workflow Mode-------')
    submit_workflow(
        parameter_dicts=[loadfn(jj) for jj in parameters],
        config_dict=load_config_file(config_file),
        work_dirs=work_dirs,
        indicated_flow_type=indicated_flow_type,
        flow_name=flow_name,
        submit_only=submit_only,
        is_debug=is_debug,
    )
    print('Completed!')
